#
#% $Id$ 
#
#
# Copyright (C) 2002-2007
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

__all__ = [ "Cell","CellBlock","Page","Workspace","Services" ];

from MeqGUI.Grid import Services

from .Services import Floater,addDataItem,removeDataItem,updateDataItem
from MeqGUI.Grid.Cell import Cell
from MeqGUI.Grid.CellBlock import CellBlock
from MeqGUI.Grid.Page import Page
from MeqGUI.Grid.Workspace import Workspace

from MeqGUI.Grid.Debug import *



class Error (RuntimeError):
  """Grid widget manipulation error""";
  _classname = "Grid.Error";

# ====== Grid.DataItem ==========================================================
# represents a per-cell data item 
class DataItem (object):
  """Represents a DataItem that is displayed in by the workspace (though not necessarily
  in a cell, or in a single GridCell.) This is meant to be constructed by data sources, 
  and passed to the workspace for displaying. Besides everything passed into the constructor,
  the dataitem also has the following attributes:
    refresh_func: the refresh function.
    viewer:     viewer class for this item (selected automatically if not specified)
    viewer_obj: instance of viewer object, once this item is displayed 
  """;
  def __init__ (self,udi,name=None,caption=None,desc=None,data=None,datatype=None,
                refresh=None,viewer=None,viewopts={}):
    """the constructor initializes standard attributes:
    udi:      the Uniform Data ID (e.g. "nodestate/<nodename>")
    name:     the name of the data item (plain text, e.g., the node name)
    caption:  caption (Qt RichText tags allowed)
    desc:     detailed description (e.g. for popup hints, etc.)
    data:     content, None if not yet initialized
    datatype: if data=None, may be used to specify which data type it will be
    refresh:  refresh function. (If not None, then the GridCell will have a 
              'refresh' button calling this function when pressed)
    viewer:   If None, a viewer will be selected from among the registered
              viewers for the data type/udi. Otherwise, provide a callable 
              viewer plug-in, or a name. See registerViewer() for details.
    viewopts: dict of dicts of viewer options (passed in as keyword arguments
              to the viewer constructor).
              viewopts[None]   applies to all viewers,
              viewopts[class]  applies to a specific class (overrides [None])
    """;
    # copy-constructor form?
    if isinstance(udi,DataItem):
      for attr in ("udi","name","caption","desc","data","refresh_func",
                   "viewer","viewopts","viewer_list"):
        setattr(self,attr,getattr(udi,attr));
    else: # else new item
      if refresh and not callable(refresh):
        raise ValueError('refresh argument must be a callable');
      self.udi      = udi;
      self.name     = name or udi;
      self.caption  = caption or udi;
      self.desc     = desc or self.name;
      self.data     = data;
      if viewopts is None:
        viewopts = {};
      self.viewopts = viewopts;
      self.refresh_func = refresh;
      # build list of compatible viewers
      self.viewer_list = Services.getViewerList(datatype or data,udi=udi);
      # if viewer specified by string, try to look it up
      if isinstance(viewer,str):
        vc = Services.getViewerByName(viewer);
        if vc:
          viewer = vc;
        else:
          raise TypeError("unknown viewer type "+viewer);
      # if viewer not specified, try to select from list
      if viewer is None:
        if not self.viewer_list:
          raise TypeError("no viewers registered and none specified");
        viewer = self.viewer_list[0];
      else:
        if not callable(viewer):
          raise TypeError('viewer argument must be a callable');
        # prepend to list
        if viewer not in self.viewer_list:
          self.viewer_list.insert(0,viewer);
      self.viewer = viewer;
      self.viewer_name = getattr(viewer,'viewer_name',viewer.__name__);
    # init other internal state
    self.viewer_obj = None;
  def __del__ (self):
    self.cleanup();
  def cleanup (self):
    _dprint(2,self.udi);
    try: self.viewer_obj.cleanup();
    except: pass;
    self.viewer_obj = None;
  def refresh (self):
    self.refresh_func and self.refresh_func();
  def is_mutable (self):
    return self.refresh_func is not None;
  def update (self,data):
    self.data = data;
    if self.viewer_obj:
      _dprint(3,'updating',self.viewer_obj,'with item',self);
      self.viewer_obj.set_data(self); 
  def highlight (self,color=True):
    if self.viewer_obj:
      self.viewer_obj.highlight(color);
  # returns True if the specified viewer class is already displaying this item
  def is_viewed_by (self,viewer):
    for v in self.viewers:
      if v is viewer or v.__class__ == viewer:
        return True;
    return False;




#usr/bin/env python
# -*- coding: utf-8 -*-

# a class to generate control buttons etc for use in controlling N-dimensional
# displays

#% $Id: ResultsRange.py 6836 2009-03-05 18:55:17Z twillis $ 

#
# Copyright (C) 2002-2007
# ASTRON (Netherlands Foundation for Research in Astronomy)
# and The MeqTree Foundation
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands, seg@astron.nl
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#  (c) 2013.				 (c) 2011.
#  National Research Council		 Conseil national de recherches
#  Ottawa, Canada, K1A 0R6 		 Ottawa, Canada, K1A 0R6
#
#  This software is free software;	 Ce logiciel est libre, vous
#  you can redistribute it and/or	 pouvez le redistribuer et/ou le
#  modify it under the terms of	         modifier selon les termes de la
#  the GNU General Public License	 Licence Publique Generale GNU
#  as published by the Free		 publiee par la Free Software
#  Software Foundation; either	 	 Foundation (version 3 ou bien
#  version 2 of the License, or	 	 toute autre version ulterieure
#  (at your option) any later	 	 choisie par vous).
#  version.
#
#  This software is distributed in	 Ce logiciel est distribue car
#  the hope that it will be		 potentiellement utile, mais
#  useful, but WITHOUT ANY		 SANS AUCUNE GARANTIE, ni
#  WARRANTY; without even the	 	 explicite ni implicite, y
#  implied warranty of			 compris les garanties de
#  MERCHANTABILITY or FITNESS FOR	 commercialisation ou
#  A PARTICULAR PURPOSE.  See the	 d'adaptation dans un but
#  GNU General Public License for	 specifique. Reportez-vous a la
#  more details.			 Licence Publique Generale GNU
#  					 pour plus de details.
#
#  You should have received a copy	 Vous devez avoir recu une copie
#  of the GNU General Public		 de la Licence Publique Generale
#  License along with this		 GNU en meme temps que ce
#  software; if not, contact the	 logiciel ; si ce n'est pas le
#  Free Software Foundation, Inc.	 cas, communiquez avec la Free
#  at http://www.fsf.org.		 Software Foundation, Inc. au
#						 http://www.fsf.org.
#
#  email:				 courriel:
#  business@hia-iha.nrc-cnrc.gc.ca	 business@hia-iha.nrc-cnrc.gc.ca
#
#  National Research Council		 Conseil national de recherches
#      Canada				    Canada
#  Herzberg Institute of Astrophysics	 Institut Herzberg d'astrophysique
#  5071 West Saanich Rd.		 5071 West Saanich Rd.
#  Victoria BC V9E 2E7			 Victoria BC V9E 2E7
#  CANADA					 CANADA
#
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import sys

from qwt.qt.QtGui import (QApplication, QMainWindow, QDialog, QGridLayout,QHBoxLayout,
         QLabel, QSizePolicy, QSlider, QPushButton, QVBoxLayout, QSpinBox, QSpacerItem, QTabWidget,QDoubleSpinBox)
from qwt.qt.QtGui import QPen, QColor,QWidget, QImage, qRgba, QFont, QFontInfo, QMenu, QActionGroup, QAction
from qwt.qt.QtCore import Qt, QSize, QObject, pyqtSignal

from MeqGUI.Plugins.BufferSizeDialog_qt5 import BufferSizeDialog


# The ResultsRange class is directly adapted from the Qt/PyQt 
# tutorial code examples.
#
# It creates a simple spinbox and slider that is used to select a data
# set for display with the result plotter. It is rather similar to the
# AxisRange class used with the ND-Controller, but the layout is
# different.

results_range_help = \
'''
This widget displays a spinbox and a slider. The spinbox and the slider both have the same value. Changing one or the other causes an event to be sent to the visualization system with a request to change the array index of the displayed data. The visualization display is then updated. The spinbox is wrapped around so that one can go directly from the highest index to the lowest.<br><br>

When you click in the area of the widget with the right mouse button a context menu will be shown. In the case of the <b>3-dimensional display</b>, this context menu will allow you to select a new active plane - one of the X, Y or Z axes. It also gives you the possibility of switching back to the 2-dimensional display. In the case of the results <b>history sequence</b> display, the menu offers you the single option of changing the number of sequential records that can be stored. If you select this option, a small dialog will appear that allows you to modify the allowable number of records that can be stored. The default is 10.<br><br> 
'''

class ResultsRange(QWidget):

#   result_index = pyqtSignal(float)
    result_index = pyqtSignal('PyQt_PyObject')
    X_axis_selected = pyqtSignal()
    Y_axis_selected = pyqtSignal()
    Z_axis_selected = pyqtSignal()
    align_camera = pyqtSignal()
    twoD_display_requested = pyqtSignal()
    save_display = pyqtSignal()
    postscript_requested = pyqtSignal()
    doubleValueChanged = pyqtSignal(float)
    update_requested = pyqtSignal()
    show_ND_Controller = pyqtSignal()
    update_scale = pyqtSignal()
    set_auto_scaling = pyqtSignal()
    adjust_results_buffer_size = pyqtSignal(int)
    display_summary_plot = pyqtSignal()


    def __init__(self, parent=None, name="",horizontal=False,draw_scale=True,hide_slider=False, use_int=False, add_spacer=True):
#     QWidget.__init__(self, parent, name)
      QWidget.__init__(self, parent)

      self.menu_table = {
      'Adjust results buffer size': 301,
      'Display summary plot': 302,
      'X Axis': 303,
      'Y Axis': 304,
      'Z Axis': 305,
      'Show 2D Display': 306,
      'Update': 307,
      'Toggle ND Controller': 308,
      'Print to Postscript file': 309,
      'Align Camera': 310,
      'Reset Auto Scaling': 312,
      'Save Display in PNG Format': 313,
      }

      self.horizontal = horizontal
      self.hide_slider = hide_slider
      self.use_int = use_int
      self.add_spacer = add_spacer
      self.allow_emit = False
      self.allow_summary = False
      self.summary_request = True
      self.toggle_ND_Controller = 1
      self.toggle_scale_display = False
      self.draw_scale = draw_scale
      self.menu = None
      self.maxVal = 100
      self.minVal = 0
      self.scaler = 1
      self.label_info = QLabel('', self)
      self.string_info =  ' '
      self.offset_index = -1
      if self.horizontal or self.use_int:
        self.spinbox = QSpinBox(self)
      else:
        self.spinbox = QDoubleSpinBox(self)
        self.spinbox.setSingleStep(0.1)
      self.spinbox.setMinimum(self.minVal)
      self.spinbox.setMaximum(self.maxVal)

      if self.horizontal:
        self.slider = QSlider(Qt.Horizontal, self)
        if self.draw_scale:
          self.slider.setTickPosition(QSlider.TicksBelow)
          self.slider.setTickInterval(self.minVal)
        self.slider.setRange(self.minVal, self.maxVal)
        self.slider.valueChanged[int].connect(self.update_slider)
        self.spinbox.valueChanged[int].connect(self.update_spinbox)
      else:
         #print('creating standard vertical QSlider')
         self.slider = QSlider(Qt.Vertical, self)
         if not self.use_int:
           self.scaler = 10000000
         min = self.scaler * self.minVal
         max = self.scaler * self.maxVal
         if self.draw_scale:
           self.slider.setTickPosition(QSlider.TicksRight)
           step = int((max-min) /20)
           #print('slider step should me ', step)
           self.slider.setTickInterval(step)
         #print('slider min and max', min,max)
         self.slider.setRange(min, max)
         self.slider.valueChanged[int].connect(self.update_slider)

        # we may not want a slider - e.g. when selecion tab pages in collections plotter
      self.spinbox.setWrapping(True)
      #print('wrapping should be set')

      if self.hide_slider:
        self.slider.hide()

      if self.horizontal or self.use_int:
        self.spinbox.valueChanged[int].connect(self.update_spinbox)
      else:
        self.spinbox.valueChanged[float].connect(self.update_spinbox)

      if self.horizontal:
        self.label_info1 = QLabel('          ', self)
        self.layout = QHBoxLayout(self)
        if self.add_spacer:
          spacer = QSpacerItem(22,9,QSizePolicy.Expanding,QSizePolicy.Minimum)
          self.layout.addItem(spacer)
        self.layout.addWidget(self.label_info)
        self.layout.addWidget(self.spinbox)
        self.layout.addWidget(self.label_info1)
        self.layout.addWidget(self.slider)
        self.setValue()
      else:
        self.layout = QVBoxLayout(self)
        self.layout.addWidget(self.label_info)
        self.layout.addWidget(self.slider)
        self.layout.addWidget(self.spinbox)
        if self.add_spacer:
          spacer = QSpacerItem(9,22,QSizePolicy.Minimum,QSizePolicy.Expanding)
          self.layout.addItem(spacer)
        self.setValue(0,reset_auto=True)

# add on-line help
      self.setWhatsThis(results_range_help)

    def setLabel(self, string_value= '', align=None):
      """ set current displayed label """
      self.label_info.setText(string_value + self.string_info);
      if align is not None:
        self.label_info.setAlignment(align);

    def setStringInfo(self, string_value= ''):
      """ assign a default leading string """
      self.string_info = string_value

    def set_decimals(self,step):
      if step < 0.1:
        self.spinbox.setDecimals(2)
      if step < 0.01:
        self.spinbox.setDecimals(3)
      if step < 0.001:
        self.spinbox.setDecimals(4)
      if step < 0.0001:
        self.spinbox.setDecimals(5)
      if step < 0.0001:
        self.spinbox.setDecimals(5)


    def setMinValue(self, min_value=0):
      """ reset allowed minimum value for spinbox and slider """
      #print('in SetMinValue with min_value', min_value)
      self.minVal = min_value
      if self.horizontal or self.use_int:
        self.spinbox.setMinimum(self.minVal)
      else:
        step = (self.maxVal-self.minVal)/20.0
        self.set_decimals(step)
        self.spinbox.setSingleStep(step)
        self.spinbox.setRange(self.minVal,self.maxVal)
      min = self.scaler * self.minVal
      max = self.scaler * self.maxVal
      #print('in setMinValue slider range set to ', min,max)
      self.slider.setRange(min, max)

    def setMaxValue(self, max_value= 0, allow_shrink=True):
      """ reset allowed maximum value for spinbox and slider """
      #print('in setMaxValue with max_value and self.maxVal`', max_value, self.maxVal)
      #print('in setMaxValue with allow_shrink = `',allow_shrink)
      if max_value < self.maxVal: 
        if allow_shrink:
          #print('we can shrink')
          self.maxVal = max_value
          min = self.scaler * self.minVal
          max = self.scaler * self.maxVal
          #print('in setMaxValue slider range set to ', min,max)
          self.slider.setRange(min, max)
          if self.horizontal or self.use_int:
            self.spinbox.setMaximum(self.maxVal)
# adding  this next line
            #print('resetting spinbox range to ', min, max)
            self.spinbox.setRange(min, max)
          else:
            self.spinbox.setSingleStep((self.maxVal-self.minVal)/20.0)
            self.spinbox.setRange(self.minVal, self.maxVal)
      else:
        if max_value > self.maxVal:
          self.maxVal = max_value
          min = self.scaler * self.minVal
          max = self.scaler * self.maxVal
          self.slider.setRange(min , max)
          #print('in setMaxValue slider range set to ', min,max)
          if self.horizontal or self.use_int:
            self.spinbox.setMaximum(self.maxVal)
          else:
            step = (self.maxVal-self.minVal)/20.0
            self.set_decimals(step)
            self.spinbox.setSingleStep(step)
            self.spinbox.setRange(self.minVal, self.maxVal)
      self.slider.setValue(self.scaler*max_value)
      self.spinbox.setValue(max_value)

    def setValue(self, value= 0, reset_auto=False):
      """ set current values shown in spinbox and slider """
      #print('in setValue raw value is  ', value)
      max = self.scaler*value
      self.slider.setValue(max)
      #print('in setValue spinbox value set to ', max)
      self.spinbox.setValue(value)
      self.initContextmenu(reset_auto)

    def setRange(self, range_value, update_value = True):
      """ define range of values shown with slider """
      #print('in setRange range value is  ', range_value)
      if range_value <= self.maxVal:
        min = self.scaler * self.minVal
        max = self.scaler * range_value
        #print('in setRange slider range set to ', max)
        self.slider.setRange(min, max)
        if self.draw_scale:
          step = int((max-min) /20)
          #print('slider step should me ', step)
          self.slider.setTickInterval(step)
        if self.horizontal or self.use_int:
          #print('spinbox max range should be', range_value)
          self.spinbox.setMaximum(range_value)
        else:
          step = (range_value-self.minVal)/20.0
          self.set_decimals(step)
          self.spinbox.setSingleStep(step)
          self.spinbox.setRange(self.minVal,range_value)
        if update_value:
          self.setValue(range_value)

    def update_slider(self, raw_slider_value):
      """ update spinbox value as function of slider value """
      #print('raw incoming slider value', raw_slider_value)
      if self.horizontal or self.use_int:
        slider_value = raw_slider_value 
        self.spinbox.setValue(int(slider_value))
      else:
        slider_value = float(raw_slider_value) / self.scaler  
        #print('update slider value', slider_value)
        self.spinbox.setValue(slider_value)

    def update_spinbox(self, spin_value):
      """ update displayed contents of spinbox """
      #print('incoming spin value', spin_value)
      self.slider.setValue(int(self.scaler* spin_value))
      if self.allow_emit:
        #print 'emitting result_index signal ', spin_value + self.offset_index
        self.result_index.emit(spin_value + self.offset_index)

    def set_emit(self, permission=True):
      """ flag to allow emitting of Qt signals """
      self.allow_emit = permission

    def X_Axis_Selected(self):
      """ emit signal to select X Axis for 3D display """
      if self.allow_emit:
#       print 'emitting X_axis_selected signal'
        self.X_axis_selected.emit(True)

    def Y_Axis_Selected(self):
      """ emit signal to select Y Axis for 3D display """
      if self.allow_emit:
        self.Y_axis_selected.emit(True)

    def Z_Axis_Selected(self):
      """ emit signal to select Z Axis for 3D display """
      if self.allow_emit:
        self.Z_axis_selected.emit(True)

    def align_camera(self):
      """ emit signal to align camera to current axis """
      #print('in align_camara')
      if self.allow_emit:
        self.align_camera.emit(True)


    def request_2D_display(self):
      """ emit signal to request 2D display """
      if self.allow_emit:
        self.twoD_display_requested.emit(True)

    def request_PNG_file(self):
      """ emit signal to request PNG printout """
      if self.allow_emit:
        self.save_display.emit(True)

    def request_postscript(self):
      """ emit signal to request Postscript printout """
      if self.allow_emit:
        self.postscript_requested.emit(True)

    def requestUpdate(self):
      """ emit signal to request update to array (for testing) """
      if self.allow_emit:
        self.update_requested.emit(True)

    def toggle_ND_controller(self):
      """ emit signal to toggle ND Controller on or off """
      toggle_id = self.menu_table['Toggle ND Controller']
      if self.toggle_ND_Controller == 1:
        self.toggle_ND_Controller = 0
        self._toggle_nd_controller.setText('Show ND Controller')
      else:
        self.toggle_ND_Controller = 1
        self._toggle_nd_controller.setText('Hide ND Controller')
      if self.allow_emit:
        self.show_ND_Controller.emit(self.toggle_ND_Controller)

    def set_summary(self, summary=True):
      """ override default value for allowing summary plot """
      self.allow_summary = summary
      toggle_id = self.menu_table['Display summary plot']
      self.menu.setItemVisible(toggle_id, self.allow_summary)

    def set_offset_index(self, offset):
      """ override default value for offset index """
      self.offset_index = offset

    def setTickInterval(self, tick_interval):
      """ override default tick interval for slider """
      if self.horizontal or self.use_int:
        self.slider.setTickInterval(tick_interval)
      else:
        self.slider.setTickInterval(self.scaler*tick_interval)

    def handleAutoScaling(self):
      """ emit signal to request 2D display """
      if self.allow_emit:
        self.set_auto_scaling.emit(True)
        self.setValue(self.maxVal,reset_auto=True)

    def handle_menu_request(self, menuid):
      """ handle requested menu option """
      #print(' handling menu request with id ', menuid)
      if menuid == self.menu_table['Reset Auto Scaling']:
        self.handleAutoScaling()
      elif menuid == self.menu_table['Display summary plot']:
        self.requestSummary()

    def initContextmenu(self, reset_auto = False):
      """Initialize the result buffer context menu """
      if self.menu is None:
        self.menu = QMenu(self)
        self.menu.triggered.connect(self.handle_menu_request)
        toggle_id = self.menu_table['Reset Auto Scaling']
        self._reset_auto_scaling = QAction('Reset Auto Scaling',self)
        self._reset_auto_scaling.setData(toggle_id)
        self.menu.addAction(self._reset_auto_scaling)
        if reset_auto:
          self._reset_auto_scaling.setVisible(True)
        else:
          self._reset_auto_scaling.setVisible(False)

        toggle_id = self.menu_table['Adjust results buffer size']
        self._adjust_results_buffer_size = QAction('Adjust results buffer size',self)
        self._adjust_results_buffer_size.setData(toggle_id)
        self.menu.addAction(self._adjust_results_buffer_size)
        self._adjust_results_buffer_size.triggered.connect(self.handleBufferSize)
        if reset_auto:
          self._adjust_results_buffer_size.setVisible(False)
        else:
          self._adjust_results_buffer_size.setVisible(True)

# option for summary plot
        toggle_id = self.menu_table['Display summary plot']
        self._display_summary_plot = QAction('Display summary plot',self)
        self.menu.addAction(self._display_summary_plot)
        self._display_summary_plot.setVisible(False)
        self._display_summary_plot.setData(toggle_id)

# options for 3D Display
        toggle_id = self.menu_table['X Axis']
        self._x_axis = QAction('X Axis',self)
        self.menu.addAction(self._x_axis)
        self._x_axis.setVisible(False)
        self._x_axis.setData(toggle_id)
        self._x_axis.triggered.connect(self.X_Axis_Selected)

        toggle_id = self.menu_table['Y Axis']
        self._y_axis = QAction('Y Axis',self)
        self.menu.addAction(self._y_axis)
        self._y_axis.setVisible(False)
        self._y_axis.setData(toggle_id)
        self._y_axis.triggered.connect(self.Y_Axis_Selected)

        toggle_id = self.menu_table['Z Axis']
        self._z_axis = QAction('Z Axis',self)
        self.menu.addAction(self._z_axis)
        self._z_axis.setVisible(False)
        self._z_axis.setData(toggle_id)
        self._z_axis.triggered.connect(self.Z_Axis_Selected)

        toggle_id = self.menu_table['Align Camera']
        self._align_camera = QAction('Align Camera',self)
        self.menu.addAction(self._align_camera)
        self._align_camera.setVisible(False)
        self._align_camera.setData(toggle_id)
        self._align_camera.setText('Align Camera to Current Axis')
        self._align_camera.triggered.connect(self.align_camera)

        toggle_id = self.menu_table['Show 2D Display']
        self._show_2d_display = QAction('Show 2D Display',self)
        self.menu.addAction(self._show_2d_display)
        self._show_2d_display.setVisible(False)
        self._show_2d_display.setData(toggle_id)
        self._show_2d_display.triggered.connect(self.request_2D_display)

        toggle_id = self.menu_table['Toggle ND Controller']
        self._toggle_nd_controller = QAction('Toggle ND Controller',self)
        self.menu.addAction(self._toggle_nd_controller)
        self._toggle_nd_controller.setVisible(False)
        self._toggle_nd_controller.setData(toggle_id)
        self._toggle_nd_controller.setText('Hide ND Controller')
        self._toggle_nd_controller.triggered.connect(self.toggle_ND_controller)

        toggle_id = self.menu_table['Update']
        self._update = QAction('Update',self)
        self.menu.addAction(self._update)
        self._update.setVisible(False)
        self._update.setData(toggle_id)
        self._update.triggered.connect(self.requestUpdate)

        toggle_id = self.menu_table['Print to Postscript file']
        self._print_to_postscript_file = QAction('Print to Postscript file',self)
        self.menu.addAction(self._print_to_postscript_file)
        self._print_to_postscript_file.setVisible(False)
        self._print_to_postscript_file.setData(toggle_id)
        self._print_to_postscript_file.triggered.connect(self.request_postscript)

        toggle_id = self.menu_table['Save Display in PNG Format']
        self._save_display_in_png_format = QAction('Save Display in PNG Format',self)
        self.menu.addAction(self._save_display_in_png_format)
        self._save_display_in_png_format.setVisible(False)
        self._save_display_in_png_format.setData(toggle_id)
        self._save_display_in_png_format.triggered.connect(self.request_PNG_file)

    def hideNDControllerOption(self):
      """ do not allow the user to toggle ND Controller """
      self._toggle_nd_controller.setVisible(False)

    def setXMenuLabel(self, text):
      """ update X axis context menu label """
      self._x_axis.setText(text)

    def setYMenuLabel(self, text):
      """ update Y axis context menu label """
      self._y_axis.setText(text)

    def setZMenuLabel(self, text):
      """ update Z axis context menu label """
      self._z_axis.setText(text)

    def disableContextmenu(self):
      """ delete the result buffer context menu """
      if not self.menu is None:
          self.menu.setParent(QWidget())
          self.menu = None

    def init3DContextmenu(self):
      """add 3D options to context menu """
      if self.menu is None:
        self.initContextmenu()
# display options for 3D Display
      self._x_axis.setVisible(True)
      self._y_axis.setVisible(True)
      self._z_axis.setVisible(True)
      self._align_camera.setVisible(True)
      self._show_2d_display.setVisible(True)
      self._toggle_nd_controller.setVisible(True)
      self._print_to_postscript_file.setVisible(True)
      self._save_display_in_png_format.setVisible(True)
      self._adjust_results_buffer_size.setVisible(False)

    def initWarpContextmenu(self):
      """add options for warped surface to context menu """
      if self.menu is None:
        self.initContextmenu()
# display options for 3D Display
      self._x_axis.setVisible(False)
      self._y_axis.setVisible(False)
      self._z_axis.setVisible(False)
      self._align_camera.setVisible(False)
      self._show_2d_display.setVisible(True)
      self._toggle_nd_controller.setVisible(True)
      self._print_to_postscript_file.setVisible(True)
      self._save_display_in_png_format.setVisible(True)
      self._toggle_vtk_scale.setVisible(True)

      self._adjust_results_buffer_size.setVisible(False)

    def displayUpdateItem(self):
      self._update.setVisible(True)

    def setResultsBuffer(self, result_value):
      """ redefine the allowable maximum number of values """
      if result_value < 0:
        return
      self.maxVal = result_value
      self.slider.setRange(self.minVal, self.maxVal)
      if self.horizontal or self.use_int:
        self.slider.setTickInterval((self.maxVal - self.minVal) / 10)
      else:
        self.slider.setStep((self.maxVal - self.minVal) / 10)
      self.spinbox.setMaximum(self.maxVal)
      self.adjust_results_buffer_size.emit(result_value)

    def handleBufferSize(self):
      """ callback to handle 'Adjust buffer' request from the context menu """
      results_dialog = BufferSizeDialog(self.maxVal, self)
      results_dialog.return_value.connect(self.setResultsBuffer)
      results_dialog.show()

    def requestSummary(self, menuid):
      """ callback to handle 'summary plot' request from the context menu """
      if menuid == self.menu_table['Display summary plot']:
        self.display_summary_plot.emit(self.summary_request)
        if self.summary_request:
          self.summary_request = False
          self.menu.changeItem(menuid, 'Discard summary plot')
        else:
          self.summary_request = True
          self.menu.changeItem(menuid, 'Display summary plot')

    def mousePressEvent(self, e):
      if Qt.RightButton == e.button():
        e.accept()
        self.menu.popup(e.globalPos());

#   def contextMenuEvent(self,ev):
#     """ The existence of this function should cause any 
#         higher-level context menu to be ignored when the 
#         right mouse button is clicked inside the widget.
#     """
#     ev.accept()



# the following tests the ResultsRange class
def make():
    demo = ResultsRange(horizontal=False, hide_slider=False)
#   demo = ResultsRange(horizontal=True, use_int=True,hide_slider=False)
    demo.setRange(0.005)
#   demo = ResultsRange()
#   demo.setRange(20)
    demo.show()
    demo.init3DContextmenu()
    return demo

# make()

def main(args):
    app = QApplication(args)
    demo = make()
#   app.setMainWidget(demo)
    app.exec_()

# main()

# Admire
if __name__ == '__main__':
    main(sys.argv)




#
#% $Id$ 
#
#
# Copyright (C) 2002-2007
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from Timba.dmi import *
from Timba import utils
from Timba import dmi_repr
from MeqGUI.GUI.pixmaps import pixmaps
from MeqGUI import Grid
from MeqGUI.GUI.browsers import GriddedPlugin,PrecisionPopupMenu
from qt import *
from qttable import *

_dmirepr = dmi_repr.dmi_repr();

_dbg = utils.verbosity(0,name='array_browser');
_dprint = _dbg.dprint;
_dprintf = _dbg.dprintf;

class ArrayTable(QTable):
  extra_col_width  = 1;   # extra column margins, in units of a charcter's width
  extra_row_height = .05;  # extra row margin, in units of a character's height

  def __init__(self,parent,**args):
    QTable.__init__(self,parent,*args);
    font = QFont("Fixed",QFontInfo(QApplication.font()).pointSize());
    font.setFixedPitch(True);
    self.setFont(font);
    self.setSelectionMode(QTable.NoSelection);
    self._arr = None;
    self._prec = (None,'f');

  # changes content
  def set_array (self,arr):
    if not 1<=arr.ndim<=2:
      raise TypeError("illegal array dimensionality");
    self._arr = arr;
    self._rank = arr.ndim;
    _dprint(3,"rank is", self._rank);
    if self._rank == 1:   
      ncol = 1;
    else:
      ncol = arr.shape[1];
    self.reset_colsizes(ncol=ncol);
    self.setNumRows(arr.shape[0]);
    self.setNumCols(ncol);
    # self.repaint_cells();
    self.apply_colsizes();

  def reset_colsizes (self,dum=None,ncol=None):
    # this is continuously updated with the minimum column width
    # required to display each column completely
    _dprint(3,ncol,self.numCols());
    self._colwidths = [0] * (ncol or self.numCols());
    self._rowheight = 0;
    self._adjust_cols = False; # True when widths have been adjusted

  def apply_colsizes (self):
    if self._adjust_cols:
      self._adjust_cols = False;
      fm = QFontMetrics(self.font());
      rh = self._rowheight + int(fm.height()*self.extra_row_height);
      dw = int(fm.width("0")*self.extra_col_width); 
      for irow in range(self.numRows()):
        self.setRowHeight(irow,rh);
      self.setTopMargin(rh);
      for (icol,w) in enumerate(self._colwidths):
        self.setColumnWidth(icol,w+dw);

  # changes precision
  def set_precision (self,prec,format):
    self._prec = (prec,format);
    self.reset_colsizes();
    self.repaint_cells();

  def get_prec (self):
    return self._prec;

  def repaint_cells (self):
    cols = list(range(min(self.columnAt(0),0),max(self.columnAt(self.width())+1,self.numCols())));
    rows = list(range(min(self.rowAt(0),0),max(self.rowAt(self.height())+1,self.numRows())));
    self._colsizes = None; # sets paint mode for paintCell
    for row in rows:
      for col in cols:
        self.updateCell(row,col);

  # redefine paintCell method to paint on-the-fly
  def paintCell(self,painter,row,col,cr,selected):
    (txt,inline) = _dmirepr.inline_str(self._arr[(row,col)[:self._rank]],prec=self._prec);
    if txt is None:
      txt = '';
    # compute bounding box of text to work out cell sizes
    rect0 = QRect(0,0,cr.width(),cr.height());
    box = painter.boundingRect(rect0,Qt.AlignLeft,txt);
    if box.width() > self._colwidths[col]:
      self._colwidths[col] = box.width();
      self._adjust_cols = True;
    if box.height() > self._rowheight:
      self._rowheight = box.height();
      self._adjust_cols = True;
    # set color
    cg = QApplication.palette().active();
    if selected:
      painter.fillRect(rect0,QBrush(cg.highlight()));
      painter.setPen(cg.highlightedText());
    else:
      painter.fillRect(rect0,QBrush(cg.base()));
      painter.setPen(cg.text());
    # actually draw text
    painter.drawText(rect0,Qt.AlignLeft,txt);

  def resizeData(self,len):
    pass;

  # redefine paint event to keep column widths correctly adjusted  
  def paintEvent(self,ev):
    # repaint
    QTable.paintEvent(self,ev);
    # will resize row/columns as needed
    self.apply_colsizes();

class ArrayBrowser(GriddedPlugin):
  _icon = pixmaps.matrix;
  viewer_name = "Array Browser";
  def is_viewable (data):
    try: return 1 <= data.ndim <=2;
    except: return False;
  is_viewable = staticmethod(is_viewable);
  
  def __init__(self,gw,dataitem,cellspec={},**opts):
    GriddedPlugin.__init__(self,gw,dataitem,cellspec=cellspec);
    self._arr = None;
    self._tbl = ArrayTable(self.wparent());
    self.set_widgets(self.wtop(),dataitem.caption,icon=self.icon());
    if dataitem and dataitem.data is not None:
      self.set_data(dataitem);
    QObject.connect(self._tbl,PYSIGNAL("fontChanged()"),self._tbl.reset_colsizes);
    context_menu = self.cell_menu();
    if context_menu is not None:
      context_menu.insertSeparator();
      menu = PrecisionPopupMenu(context_menu,prec=self._tbl.get_prec());
      context_menu.insertItem(pixmaps.precplus.iconset(),'Number format',menu);
      QWidget.connect(menu,PYSIGNAL("setPrecision()"),\
                      self._tbl.set_precision);
      
  def wtop (self):
    return self._tbl;
  
  def set_data (self,dataitem,**opts):
    # clear everything and reset data as new
    self._tbl.set_array(dataitem.data);
    self.enable();
    self.flash_refresh();
    
# Give ArrayBrowser slightly higher priority for arrays
Grid.Services.registerViewer(array_class,ArrayBrowser,priority=15);



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#
#% $Id$
#
#
# Copyright (C) 2002-2007
# The MeqTree Foundation &
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

#ensure we use API 2 from pyqt regardless of python version with python qwt
import six
if not six.PY3:
  raise RuntimeError("Meqtrees has migrated to Python 3.x and no longer supports running Python 2.x")

import sip

sip.setapi('QString', 2)
sip.setapi('QVariant', 2)
sip.setapi('QDate', 2)
sip.setapi('QDateTime', 2)
sip.setapi('QTextStream', 2)
sip.setapi('QTime', 2)
sip.setapi('QUrl', 2)
import os
os.environ['QT_API'] = 'pyqt' #qt4 not 5 as is installed on most sytems running python 3!

debuglevels = {};
options = {};

import Timba.utils
import os
import sys
import traceback
import socket
import PyQt4.QtCore as QtCore
QtCore.QCoreApplication.setAttribute(QtCore.Qt.AA_X11InitThreads)
from PyQt4.QtCore import QCoreApplication as qca
qca.setAttribute(QtCore.Qt.AA_X11InitThreads)

# these don't like being re-imported in TDL, so try them here
try:
  import pyrap.tables
except ImportError:
  try: import pyrap_tables; 
  except ImportError: pass;
try:
  import pyrap.measures
except ImportError:
  try: import pyrap_measures; 
  except ImportError: pass;

def trace_lines (frame, event, arg):
  if event == "line":
    print(("%s:%d"%(frame.f_code.co_filename,frame.f_lineno)));
  return trace_lines;

if __name__ == "__main__":
  if not os.access('.',os.W_OK):
    print(("You do not have write permissions to your current working directory,",os.getcwd()));
    print("MeqTrees must be run in from a directory you can write to, such as your home dir.") ;
    print("Please cd to a writable directory and try again.") ;
    sys.exit(1);

  # tell verbosity class to not parse argv -- we do it ourselves here
  Timba.utils.verbosity.disable_argv();
  # parse options is the first thing we should do
  from optparse import OptionParser
  parser = OptionParser()
  parser.add_option("-p", "--port",dest="port",type="int",
                    default=4000+os.getuid(),
                    help="TCP port to listen on for connections from remote meqservers");
  parser.add_option("-s", "--socket",dest="socket",type="string",
                    default="meqbrowser-%d:1"%os.getuid(),
                    help="local Unix socket to listen on for connections from local meqservers, use None for none");
  parser.add_option("-d", "--debug",dest="debug",type="string",action="append",metavar="Context=Level",
                    help="(for debugging C++ code) sets debug level of the named C++ context. May be used multiple times.");
  parser.add_option("-v", "--verbose",dest="verbose",type="string",action="append",metavar="Context=Level",
                    help="(for debugging Python code) sets verbosity level of the named Python context. May be used multiple times.");
  parser.add_option("-t", "--trace",dest="trace",action="store_true",
                    help="(for debugging Python code) enables line tracing of Python statements");
  (options, rem_args) = parser.parse_args();

  if options.trace:
    sys.settrace(trace_lines);

  for optstr in (options.debug or []):
    opt = optstr.split("=") + ['1'];
    context,level = opt[:2];
    debuglevels[context] = int(level);

  for optstr in (options.verbose or []):
    opt = optstr.split("=") + ['1'];
    context,level = opt[:2];
    Timba.utils.verbosity.set_verbosity_level(context,int(level));

  print("Welcome to the MeqTrees Browser!");
  print("Please wait a second while the GUI starts up.");

  if not sys.platform.startswith('linux'):
    print("Removing left-over Unix socket files associated with MeqTrees")
    os.system('rm -f /tmp/\=meqbrowser-%d:*' % (os.getuid(),))
    os.system('rm -f /tmp/\=meqserver-%d:*' % (os.getuid(),))

# first things first: setup app defaults from here and from
# command line (this has to go first, as other modules being imported
# may depend on app_defaults settings)
from Timba.Apps import app_defaults
from Timba.Apps import config

from MeqGUI import qt_threading
from Timba import octopussy
from Timba.Apps import meqserver

import Timba.utils

from MeqGUI.GUI import app_proxy_gui
app_proxy_gui.mainapp();

# get version numbers
try:
  from Timba.version_info.release import release
except:
  release = '(developer build)';
try:
  from Timba.version_info.svn_revision import svn_revision
  svn_revision = "<p align='right'>(svn revision %s)</p>"%svn_revision;
except:
  svn_revision = '';

from Kittens import pixmaps;
pixmaps.load_icons('treebrowser', "Timba");
app_proxy_gui.set_splash_screen(pixmaps.redhood_300.pm,
	"Welcome to MeqTrees %s %s"%(release,svn_revision));

# ugly, but for some reason Purr objects to being imported from within TDL
if False:
  try: import Purr
  except: pass;


def importPlugin (name,location='MeqGUI.Plugins'):
  name = location + '.' +name;
  try:
    __import__(name,globals(),locals(),[]);
  except Exception as what:
    print(("\n WARNING: couldn't import plugin '%s' (%s)"%(name,what)));
    print('  This plugin will not be available.');

### import plug-ins
### disabled for now since they haven't been ported to Qt4 yet

if False:
  importPlugin('array_browser');

importPlugin('quickref_plotter');
importPlugin('collections_plotter');
importPlugin('array_plotter');
importPlugin('result_plotter');

# importPlugin('AxM.pyvis.pynode_plotter',location='Timba.Contrib')

### retired plugins
# importPlugin('pylab_plotter');
# importPlugin('node_execute');
# importPlugin('histogram_plotter');
# importPlugin('svg_plotter');
# importPlugin('history_plotter');
# importPlugin('parmfiddler');
# importPlugin('TableInspector');
# importPlugin('stream_control');

def meqbrowse (debug={},**kwargs):
  args = dict(app_defaults.args);
  args['spawn'] = False;
  for d,l in list(debug.items()):
    debuglevels[d] = max(debuglevels.get(d,0),l);

  # insert '' into sys.path, so that CWD is always in the search path
  sys.path.insert(1,'');
  if debuglevels:
    octopussy.set_debug(debuglevels);

  # start octopussy if needed
  port = options.port;
  sock = options.socket;
  if sock == "None" or sock == "none":
    sock = "";
    print("Not binding to a local socket.");
  else:
    # Use abstract socket on Linux and corresponding file-based socket elsewhere
    sock = "="+sock;
    if not sys.platform.startswith('linux'):
      sock = "/tmp/"+sock;
    print(("Binding to local socket %s"%sock));
    # check local socket
    sk = socket.socket(socket.AF_UNIX);
    try:
      sk.bind( ("\0"+sock[1:]) if sock[0] == '=' else sock);
    except:
      print(("Error binding to local socket %s"%sock));
      print("This probably means that another meqbrowser is already running.")
      print("For advanced use, see the -s option (use -h to get help).");
      sys.exit(1);
    sk.close();
    if sock[0] != '=' and os.path.exists(sock):
      os.remove(sock)
  print(("Binding to TCP port %d, remote meqservers may connect with gwpeer=<host>:%d"%(port,port)));
  if not octopussy.is_initialized():
    octopussy.init(gwclient=False,gwtcp=port,gwlocal=sock);
  if not octopussy.is_running():
    octopussy.start(wait=True);
  # start meqserver
  # print "starting meqserver"
  args['gui'] = True;
  mqs = meqserver.meqserver(**args);

#   try:
#     import psyco
#     psyco.log('psyco.log');
#     psyco.profile();
#     print "****** psyco enabled.";
#   except:
#     print "****** You do not have the psyco package installed.";
#     print "****** Proceeding anyway, things will run some what slower.";
#     pass;

  mqs.run_gui();
  mqs.disconnect();
  octopussy.stop();

if __name__ == '__main__':
  meqbrowse();

#  thread = qt_threading.QThreadWrapper(meqbrowse);
#  print 'starting main thread:';
#  thread.start();
#  thread.join();
#  print 'main thread rejoined, exiting';




#!/usr/bin/env python3

# an adaption of the PyQwt-4.2 QwtImagePlotDemo.py example so that
# we can draw color bars


#% $Id: QwtColorBar.py 6836 2009-03-05 18:55:17Z twillis $ 

#
# Copyright (C) 2002-2007
# ASTRON (Netherlands Foundation for Research in Astronomy)
# and The MeqTree Foundation
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands, seg@astron.nl
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
#  (c) 2013.				 (c) 2011.
#  National Research Council		 Conseil national de recherches
#  Ottawa, Canada, K1A 0R6 		 Ottawa, Canada, K1A 0R6
#
#  This software is free software;	 Ce logiciel est libre, vous
#  you can redistribute it and/or	 pouvez le redistribuer et/ou le
#  modify it under the terms of	         modifier selon les termes de la
#  the GNU General Public License	 Licence Publique Generale GNU
#  as published by the Free		 publiee par la Free Software
#  Software Foundation; either	 	 Foundation (version 3 ou bien
#  version 2 of the License, or	 	 toute autre version ulterieure
#  (at your option) any later	 	 choisie par vous).
#  version.
#
#  This software is distributed in	 Ce logiciel est distribue car
#  the hope that it will be		 potentiellement utile, mais
#  useful, but WITHOUT ANY		 SANS AUCUNE GARANTIE, ni
#  WARRANTY; without even the	 	 explicite ni implicite, y
#  implied warranty of			 compris les garanties de
#  MERCHANTABILITY or FITNESS FOR	 commercialisation ou
#  A PARTICULAR PURPOSE.  See the	 d'adaptation dans un but
#  GNU General Public License for	 specifique. Reportez-vous a la
#  more details.			 Licence Publique Generale GNU
#  					 pour plus de details.
#
#  You should have received a copy	 Vous devez avoir recu une copie
#  of the GNU General Public		 de la Licence Publique Generale
#  License along with this		 GNU en meme temps que ce
#  software; if not, contact the	 logiciel ; si ce n'est pas le
#  Free Software Foundation, Inc.	 cas, communiquez avec la Free
#  at http://www.fsf.org.		 Software Foundation, Inc. au
#						 http://www.fsf.org.
#
#  email:				 courriel:
#  business@hia-iha.nrc-cnrc.gc.ca	 business@hia-iha.nrc-cnrc.gc.ca
#
#  National Research Council		 Conseil national de recherches
#      Canada				    Canada
#  Herzberg Institute of Astrophysics	 Institut Herzberg d'astrophysique
#  5071 West Saanich Rd.		 5071 West Saanich Rd.
#  Victoria BC V9E 2E7			 Victoria BC V9E 2E7
#  CANADA					 CANADA
#
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from math import log
from math import exp
from math import pi
import numpy

from qwt.qt.QtGui import QApplication, QSizePolicy, QMessageBox
from qwt.qt.QtGui import QPen, QColor,QWidget, QImage, QFont, QFontInfo, QMenu, QActionGroup, QAction
from qwt.qt.QtCore import Qt, pyqtSignal, QSize, QObject
from qwt import (QwtPlot, QwtPlotMarker, QwtPlotCurve,
                 QwtPlotItem, QwtLinearColorMap,
                 QwtLogScaleEngine, QwtLinearScaleEngine,
                 QwtScaleMap, QwtScaleDraw, QwtScaleDiv)

HAS_TIMBA = False
try:
  from MeqGUI.Plugins.QwtPlotImage_qt5 import QwtPlotImage
  from MeqGUI.Plugins.QwtSpy_qt5 import Spy
  from Timba.utils import verbosity
  _dbg = verbosity(0,name='QwtColorBar');
  _dprint = _dbg.dprint;
  _dprintf = _dbg.dprintf;
  HAS_TIMBA = True
except:
  pass

# called by the QwtImagePlot class
def square(n, min, max):
    t = numpy.arange(min, max, float(max-min)/(n-1))
    #return outer(cos(t), sin(t))
    return numpy.cos(t)*numpy.sin(t)[:,numpy.newaxis]
# square()



colorbar_instructions = \
'''The colorbar displays the current range of intensities in the corresponding image display. You can interact with the colorbar to change the range of intensities displayed in the image.<br><br>
Button 1 (Left): If you click the <b>left</b> mouse button on a location inside the colorbar and then drag it, a rectangular square will be seen. When you release the left mouse button, the range of intensity defined in the vertical (Y) direction will now specify the maximum range of intensity that will be shown in the image display. Image pixels with values greater or less than the selected range will be plotted with the maximum or minimum allowed values. The color rainbow or grayscale will always cover the specified range of pixels, so you can obtain increased detail by zooming in on an intensity range.<br><br>
Button 2 (Right):Clicking the <b>right</b> mouse button in the colorbar window will cause a context menu with to appear. If you have already zoomed into the colorbar using the method described above, one of the options to appear will be 'unzoom intensity range '. If you click on this menu item, then the colorbar scale (and the image scale) is reset to the intrinsic range associated with the current image. The other option allows you to lock or unlock the colorbar. If you lock the colorbar, then the scale will be fixed; even if you load a new image, the scale will not change. Nor will you be able to reset the intensity scale to the default, if you had previously zoomed in. To enable the colorbar to change the intensity range, you must then unlock the colorbar.'''

class QwtColorBar(QwtPlot):
    menu_table = {
        'unzoom intensity range': 200,
        'lock colorbar scale': 201,
        'unlock colorbar scale': 202,
        }

    set_image_range = pyqtSignal(int,int,int,int)

    def __init__(self, colorbar_number=0, parent=None):
        QwtPlot.__init__(self, parent)
        self._mainwin = parent and parent.topLevelWidget()
        self.colorbar_number = colorbar_number
        # create copy of standard application font..
        font = QFont();
        fi = QFontInfo(font);
        # and scale it down to 70%
        font.setPointSize(fi.pointSize()*0.7);
        # apply font to QwtPlot
#       self.setTitleFont(font);
        for axis in range(0,4):
          self.setAxisFont(axis,font);
#         self.setAxisTitleFont(axis,font);

	# make a QwtPlot widget
        self.plotLayout().setCanvasMargin(0)
        self.plotLayout().setAlignCanvasToScales(1)
	# set axis parameters
        self.enableAxis(QwtPlot.yLeft)
        self.enableAxis(QwtPlot.xBottom, False)
        self.setAxisLabelRotation(QwtPlot.yLeft,270)
        self.setAxisLabelAlignment(QwtPlot.yLeft, Qt.AlignTop)
        # default color bar
        self.plotImage = QwtPlotImage(self)
        self.plotImage.attach(self)
        self.updateDisplay()
        self.min = 0.0
        self.max = 256.0
        self.is_active = False
        self.log_scale = False
        self.ampl_phase = False
        self.bar_array = numpy.reshape(numpy.arange(self.max), (1,256))
        self.y_scale = (self.min, self.max)
        self.plotImage.setData(self.bar_array, None, self.y_scale)

# Over-ride default QWT Plot size policy of MinimumExpanding
# Otherwise minimum size of plots is too large when embedded in a
# QGridlayout
        self.setSizePolicy(QSizePolicy.Expanding,QSizePolicy.Expanding)

        # width limits - the following seem reasonable
        # we don't want the bar to resize itself freely - it becomes too big!
#       self.setMaximumWidth(self.sizeHint().width() * 2.5)
        self.setMaximumWidth(self.sizeHint().width() * 1.5)

        self.spy = Spy(self.canvas())
        self.spy.MouseMove.connect(self.MouseMoveEvent)
        self.spy.MousePress.connect(self.MousePressEvent)
        self.spy.MouseRelease.connect(self.MouseReleaseEvent)

        self.zoomStack = []
        self.xzoom_loc = None
        self.yzoom_loc = None
        self.prev_xpos = None
        self.prev_ypos = None
        self.raw_image_min = None
        self.raw_image_max = None
        # create zoom curve
        self.zoom_outline = QwtPlotCurve()

        self.setMouseTracking(True)
        self.installEventFilter(self)

        # add intructions on how to use
        self.setWhatsThis(colorbar_instructions)


        # create pull_down menu and add menu components
        if self._mainwin:
          self._menu = QMenu(self._mainwin);
        else:
          self._menu = QMenu(None);

        self._unzoom_action = QAction('unzoom intensity range', self)
        self._menu.addAction(self._unzoom_action)
        self._unzoom_action.setVisible(False)
        self._unzoom_action.triggered.connect(self.unzoom)

        self._lock_colorbar = QAction('lock colorbar scale', self)
        self._menu.addAction(self._lock_colorbar)
        self._lock_colorbar.triggered.connect(self.lock_colorbar)  

        self._unlock_colorbar = QAction('unlock colorbar scale', self)
        self._menu.addAction(self._unlock_colorbar)
        self._unlock_colorbar.setVisible(False)
        self._unlock_colorbar.triggered.connect(self.unlock_colorbar)  
        self._lock_bar = False
        
# for drag & drop stuff ...
        self.setAcceptDrops(True)
        self.yhb = 0
        self.ylb = 0
        self.xhb = 0
        self.xlb = 0

    # __init__()

    def lock_colorbar(self):
      self._lock_colorbar.setVisible(False)
      self._unlock_colorbar.setVisible(True)
      if len(self.zoomStack):
        self._unzoom_action.setVisible(False)
      self._lock_bar = True
      self.emit_range()

    def unlock_colorbar(self):
      self._lock_colorbar.setVisible(True)
      self._unlock_colorbar.setVisible(False)
      if len(self.zoomStack):
        self._unzoom_action.setVisible(True)
      self._lock_bar = False
      self.emit_range()
        

    def get_data_range(self):
      """ returns range of this widget when called by 'foreign'
          widget on which we have done a drop event
      """
      rng = (self.min, self.max, self.colorbar_number, self.ampl_phase)
      return rng

    def dragEnterEvent(self, event):
      """ drag & drop event callback entered when we move out of or
          in to a widget 
      """ 
      try:
        if event.mimeData().hasText():
          event.acceptProposedAction()
      except:
        pass

    def dropEvent(self, event):
        """ callback that handles a drop event from drag & drop """
        if event.source() == self:
          return
        if event.mimeData().hasText():
          command_str = str(event.mimeData().text())
          if command_str.find('copyColorRange') > -1:
              components = event.source().get_data_range()
              min = components[0]
              max = components[1]
              colorbar = components[2] 
              amp_phas = components[3]
              if self.colorbar_number == colorbar and not self._lock_bar:
                self.zoomStack == []
                try:
                  self.zoomState = (
                    self.axisScaleDiv(QwtPlot.yLeft).lBound(),
                    self.axisScaleDiv(QwtPlot.yLeft).hBound(),
                    )
                except:
                  self.zoomState = (
                    self.axisScaleDiv(QwtPlot.yLeft).lowerBound(),
                    self.axisScaleDiv(QwtPlot.yLeft).upperBound(),
                    )
                self.zoomStack.append(self.zoomState)
                self._unzoom_action.setVisible(True)
                self.setRange(min, max, colorbar, amp_phas)
        else:
          message= 'QwtColorBar dropEvent decode failure'
          mb_reporter = QMessageBox.information(self, "QwtColorBar",message)
        event.acceptProposedAction()

    def startDrag(self,event):
      """ operations done when we start a drag event """ 
      drag = Qt.QDrag(self)
      mimedata = Qt.QMimeData()
      passed_string = 'copyColorRange'
      mimedata.setText(passed_string)
      drag.setMimeData(mimedata)
      drag.exec_()
      event.accept()

    def setRange(self, min, max,colorbar_number=0, ampl_phase=False):
      """ sets display range for this colorbar and emits signal
          to associated display to set corresponding range 
      """ 
      if ampl_phase != self.ampl_phase:
        return

      if colorbar_number == self.colorbar_number:
        if min > max:
          temp = max
          max = min
          min = temp
        if abs(max - min) < 2.0e-8:
          if max == 0.0 or min == 0.0:
            min = -0.1
            max = 0.1
          else:
            min = 0.9 * min
            max = 1.1 * max
        self.min = min * 1.0
        self.max = max * 1.0

        # send event to display_image.py that range has changed
        self.emit_range()

        if self.log_scale:
          max = log(self.max)
          min = log(self.min)
          delta = (max - min) / 255.0
          for i in range (256):
            self.bar_array[0,i] = exp (min + i * delta)
        else:
          delta = (self.max - self.min) / 255.0
          for i in range (256):
            self.bar_array[0,i] = self.min + i * delta
        self.y_scale = (self.min, self.max)
        self.plotImage.setData(self.bar_array, None, self.y_scale)
        self.show()
        self.replot()
    # set Range()

    def getTransformOffset(self):
      """ get the offset value for a plot with a log scale """
      return self.plotImage.getTransformOffset()
 
    def setScales(self):
      self.plotImage.setLogScale(self.log_scale)
      self.plotImage.setLogYScale(self.log_scale)
      self.setAxisAutoScale(QwtPlot.yLeft) 
      if self.log_scale:
        self.setAxisScaleEngine(QwtPlot.yLeft, QwtLogScaleEngine())
      else:
        self.setAxisScaleEngine(QwtPlot.yLeft, QwtLinearScaleEngine())

    def handleRangeParms(self, limits,colorbar_number,log_scale,ampl_phase):
      if ampl_phase is None:
        ampl_phase = False
      self.setMaxRange(limits, colorbar_number, log_scale, ampl_phase)

    def setMaxRange(self, limits, colorbar_number=0, log_scale=False, ampl_phase=False):
      """ sets maximum range parameters for this colorbar """
#     print('in colorbar setMaxRange : colorbar numbers ',colorbar_number,self.colorbar_number)
      if colorbar_number == self.colorbar_number:
        self.ampl_phase = ampl_phase
        
        if self.ampl_phase is None:
          self.ampl_phase = False
        self.log_scale = log_scale
        self.setScales()
        min = limits[0]
        max = limits[1]
        try:
          self.raw_image_min = limits[2]
          self.raw_image_max = limits[3]
        except:
          pass
        if min > max:
            temp = max
            max = min
            min = temp
#       if abs(max - min) < 0.00005:
        if abs(max - min) < 2.0e-8:
          if max == 0.0 or min == 0.0:
            min = -0.1
            max = 0.1
          else:
            min = 0.9 * min
            max = 1.1 * max
        if not self._lock_bar:
          self.image_min = min * 1.0
          self.image_max = max * 1.0
          self.min = self.image_min
          self.max = self.image_max
        if self.log_scale:
          if self.min <= 0.0:
            offset = -1.0 * self.min + 0.001
            self.min = self.min + offset
            self.max = self.max + offset
            self.image_min = self.min
            self.image_max = self.max
          max = log(self.max)
          min = log(self.min)
          delta = (max - min) / 255.0
          for i in range (256):
            self.bar_array[0,i] = exp (min + i * delta)
          if HAS_TIMBA:_dprint(3, 'log bar array is ',self.bar_array)
        else:
          delta = (self.max - self.min) / 255.0
          for i in range (256):
            self.bar_array[0,i] = self.min + i * delta
        if not self._lock_bar:
          self.y_scale = (self.min, self.max)
          self.plotImage.setData(self.bar_array, None, self.y_scale)
          self.show()
          self.replot()
    # setMaxRange()

    def showDisplay(self, show_self, colorbar_number=0):
      """ callback to show or hide this colorbar """
#     print('in colorbar show display with self.colorbar_number', self.colorbar_number)
      if colorbar_number == self.colorbar_number:
        self.is_active = True
        if show_self > 0:
          self.show()
#         print('should show')
        else:
          self.hide()
#         print('should hide')
        self.replot()
    # showDisplay

    def unHide(self):
      """ callback to show this colorbar """
      if self.is_active:
        self.show()


    def emit_range(self):
#     print('emitting range')
      self.set_image_range.emit(self.min, self.max, self.colorbar_number,self._lock_bar)


    def setBarLock(self, set_lock=False):
      self._lock_bar = set_lock  
      if self._lock_bar:
        toggle_id = self.menu_table['lock colorbar scale']
        self._menu.setItemVisible(toggle_id, False)
        toggle_id = self.menu_table['unlock colorbar scale']
        self._menu.setItemVisible(toggle_id, True)
      else:
        toggle_id == self.menu_table['unlock colorbar scale']
        self._menu.setItemVisible(toggle_id, False)
        toggle_id = self.menu_table['lock colorbar scale']
        self._menu.setItemVisible(toggle_id, True)
      self.emit_range()

    def unzoom(self):
      """ callback to set range of this colorbar back to default """
      if len(self.zoomStack):
        while len(self.zoomStack):
          ymin, ymax = self.zoomStack.pop()

        self._unzoom_action.setVisible(False)

      if not self.raw_image_min is None:
        self.image_min = self.raw_image_min
        self.image_max = self.raw_image_max
      self.setAxisScale(QwtPlot.yLeft, self.image_min, self.image_max)
      if self.image_min > self.image_max:
        temp = self.image_max
        self.image_max = self.image_min
        self.image_min = temp
      self.setRange(self.image_min, self.image_max, self.colorbar_number,self.ampl_phase)

    # set the type of colorbar display, can be one of "hippo", "grayscale" 
    # or "brentjens"
    def setDisplayType(self,display_type):
        self.plotImage.setDisplayType(display_type)
        self.plotImage.setData(self.bar_array, None, self.y_scale)
        self.replot()
    # setDisplayType()


    def drawCanvasItems(self, painter, rectangle, maps, filter):
        self.plotImage.drawImage(
            painter, maps[QwtPlot.xBottom], maps[QwtPlot.yLeft])
        QwtPlot.drawCanvasItems(self, painter, rectangle, maps, filter)
    # drawCanvasItems()

    def MouseMoveEvent(self, e):
      """ callback to handle mouse moved event """
#     print('in mouse move event')
      position = e.pos()
      xPos = position.x()
      yPos = position.y()
      self.xpos = self.invTransform(QwtPlot.xBottom, xPos)
      self.ypos = self.invTransform(QwtPlot.yLeft, yPos)
      if not self.xzoom_loc is None:
        self.xzoom_loc = [self.press_xpos, self.press_xpos,  self.xpos, self.xpos,self.press_xpos]
        self.yzoom_loc = [self.press_ypos, self.ypos,  self.ypos, self.press_ypos,self.press_ypos]
        if self.zoom_outline is None:
          self.zoom_outline = QwtPlotCurve()
        self.zoom_outline.setSamples(self.xzoom_loc,self.yzoom_loc)
        self.replot()

        # Test if mouse has moved outside the plot. If yes, we're 
        # starting a drag.
        if xPos < self.xlb-10 or xPos > self.xhb+10 or yPos > self.ylb+10 or yPos < self.yhb-10:
          if not self.xzoom_loc is None:
            self.zoom_outline.detach()
            self.xzoom_loc = None
            self.yzoom_loc = None
            self.replot()
          self.startDrag(e)
    # onMouseMoved()

    def MousePressEvent(self, e):
      """ callback to handle mouse pressed event """
      if Qt.LeftButton == e.button():
        # get bounds of plot. Keep them around for later test if
        # we're initiating a drag operation
        try:
          self.yhb = self.transform(QwtPlot.yLeft, self.axisScaleDiv(QwtPlot.yLeft).hBound())
          self.ylb = self.transform(QwtPlot.yLeft, self.axisScaleDiv(QwtPlot.yLeft).lBound())
          self.xhb = self.transform(QwtPlot.xBottom, self.axisScaleDiv(QwtPlot.xBottom).hBound())
          self.xlb = self.transform(QwtPlot.xBottom, self.axisScaleDiv(QwtPlot.xBottom).lBound())
        except:
          self.yhb = self.transform(QwtPlot.yLeft, self.axisScaleDiv(QwtPlot.yLeft).upperBound())
          self.ylb = self.transform(QwtPlot.yLeft, self.axisScaleDiv(QwtPlot.yLeft).lowerBound())
          self.xhb = self.transform(QwtPlot.xBottom, self.axisScaleDiv(QwtPlot.xBottom).upperBound())
          self.xlb = self.transform(QwtPlot.xBottom, self.axisScaleDiv(QwtPlot.xBottom).lowerBound())
        # Python semantics: self.pos = e.pos() does not work; force a copy
        if HAS_TIMBA:_dprint(3, 'e.pos() ', e.pos())
        self.press_xpos = self.xpos 
        self.press_ypos = self.ypos
        if HAS_TIMBA:_dprint(3, 'self.xpos self.ypos ', self.xpos, ' ', self.ypos)
        if not self._lock_bar:
          self.xzoom_loc = [self.press_xpos]
          self.yzoom_loc = [self.press_ypos]
          self.zoom_outline.attach(self)
          if self.zoomStack == []:
            try:
              self.zoomState = (
                self.axisScaleDiv(QwtPlot.yLeft).lBound(),
                self.axisScaleDiv(QwtPlot.yLeft).hBound(),
                )
            except:
              self.zoomState = (
                self.axisScaleDiv(QwtPlot.yLeft).lowerBound(),
                self.axisScaleDiv(QwtPlot.yLeft).upperBound(),
                )
          # fake a mouse move to show the cursor position
#         self.onMouseMoved(e)
      elif Qt.RightButton == e.button():
        e.accept()
        self._menu.popup(e.globalPos())
    # onMousePressed()

    def MouseReleaseEvent(self, e):
      """ handles mouse release event if we're not doing a drop """

      # if color bar is locked, do nothing
      if self._lock_bar:
        return

      if Qt.LeftButton == e.button():
        xmin = min(self.xpos, self.press_xpos)
        xmax = max(self.xpos, self.press_xpos)
        ymin = min(self.ypos, self.press_ypos)
        ymax = max(self.ypos, self.press_ypos)
        if not self.xzoom_loc is None:
          self.zoom_outline.detach()
          self.xzoom_loc = None
          self.yzoom_loc = None
        if xmin == xmax or ymin == ymax:
          return
        self.zoomStack.append(self.zoomState)
#       print('self.zoomsSack', self.zoomStack)
        self.zoomState = (ymin, ymax)
        if len(self.zoomStack):
          self._unzoom_action.setVisible(True)
      elif Qt.RightButton == e.button():
        if len(self.zoomStack):
          ymin, ymax = self.zoomStack.pop()
        else:
          return

      self.setAxisScale(QwtPlot.yLeft, ymin, ymax)
      if ymin > ymax:
        temp = ymax
        ymax = ymin
        ymin = temp
      self.setRange(ymin, ymax, self.colorbar_number, self.ampl_phase)

      self.replot()
    # onMouseReleased()

    def updateDisplay(self):
      # calculate 3 NumPy arrays
      self.gain = 1.0
      # image
      self.plotImage.setData(
            square(512,-1.0 * self.gain*pi, self.gain*pi), (-1.0*self.gain*pi, self.gain*pi), (-1.0*self.gain*pi, self.gain*pi))


# class QwtColorBar


# the following tests the QwtColorBar class
def make():
    demo = QwtColorBar()
#   demo.setMaxRange((1.0e-11, 0.001),0,True)
#   demo.setMaxRange((1.0e-11, 1.0),0)
    demo.setMaxRange((-1.0,1.0), colorbar_number=0, log_scale=True)

#   demo.updateDisplay()
#   demo.resize(50, 200)
    demo.show()
    demo.replot()
    return demo

# make()

def main(args):
    app = QApplication(sys.argv)
    demo = make()
    app.exec_()

# main()

# Admire
if __name__ == '__main__':
    import sys
    main(sys.argv)







#!/usr/bin/env python3

import os
import sys
from distutils.core import setup
from distutils.command.install import INSTALL_SCHEMES

try:
    import sip
    if not hasattr(sip, 'setapi'):
        raise ImportError("SIP not installed or does not have the setapi attribute")
except ImportError:
    print("Cannot import SIP. This is not available from PyPI and has "
          "to be installed from your distribution streams")
    sys.exit(1)

try:
    import PyQt4
    import PyQt4.Qt
except ImportError:
    print("Cannot import PyQt4. This is not available from PyPI and has "
          "to be installed from your distribution streams")
    sys.exit(1)

try:
    pass
except ImportError:
    print("Cannot import qwt. The package is not available from PyPI and has "
          "to be installed from your distribution streams")

def fullsplit(path, result=None):
    """
    Split a pathname into components (the opposite of os.path.join) in a
    platform-neutral way.
    """
    if result is None:
        result = []
    head, tail = os.path.split(path)
    if head == '':
        return [tail] + result
    if head == path:
        return result
    return fullsplit(head, [tail] + result)

# Tell distutils not to put the data_files in platform-specific installation
# locations. See here for an explanation:
# http://groups.google.com/group/comp.lang.python/browse_thread/thread/35ec7b2fed36eaec/2105ee4d9e8042cb
for scheme in list(INSTALL_SCHEMES.values()):
    scheme['data'] = scheme['purelib']

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

packages = []
data_files = []
scripts = ["scripts/meqbrowser.py"]
root_dir = os.path.dirname(__file__)
if root_dir != '':
    os.chdir(root_dir)

for dirpath, dirnames, filenames in os.walk('MeqGUI'):
    # Ignore dirnames that start with '.'
    dirnames[:] = [d for d in dirnames if not d.startswith('.') and d != '__pycache__']
    if '__init__.py' in filenames:
        packages.append('.'.join(fullsplit(dirpath)))
    elif filenames:
        data_files.append([dirpath, [os.path.join(dirpath, f) for f in filenames]])

install_requires = [
    'numpy',
    'python-casacore',
    'six',
    'configparser',
    'matplotlib',
    'PythonQwt>=0.10.1'
] + [ # meqtrees sister packages    
    'purr',
    'astro-kittens',
    'meqtrees-cattery',
    'astro-tigger-lsm',
    'owlcat',
]

setup(name='meqtrees-frontend',
      version='1.0.0',
      python_requires='>=3.6.0',
      description='MeqTrees-based frameworks for simulation and calibration of radio interferometers ',
      author='Oleg Smirnov',
      author_email='osmirnov@gmail.com',
      url='https://github.com/ska-sa/meqtrees-cattery',
      packages=packages,
      scripts=scripts,
      data_files=data_files,
      install_requires=install_requires,
      long_description=long_description,
      long_description_content_type='text/markdown'
     )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#
#% $Id$ 
#
#
# Copyright (C) 2002-2007
# The MeqTree Foundation & 
# ASTRON (Netherlands Foundation for Research in Astronomy)
# P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>,
# or write to the Free Software Foundation, Inc., 
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

from Timba.dmi import *
from Timba.utils import *
from MeqGUI.GUI.pixmaps import pixmaps
from MeqGUI.GUI.widgets import *
from MeqGUI.Grid.Debug import *
import MeqGUI.Grid.Cell
import MeqGUI.Grid.Page
from Timba import *
from MeqGUI import GUI, Plugins, Grid

import weakref
import re
import gc
import types

from PyQt4.Qt import *
from Kittens.widgets import PYSIGNAL

_reg_viewers = {};
_reg_viewers_byname = {};

# registers a viewer plug-in for the specified type
#
# a viewer plug-in must provide the following interface:
#

def registerViewer (tp,viewer,check_udi=lambda x:True,priority=0):
  """Registers a viewer for the specified type:
  The 'viewer' argument must be a class (or callable) providing the following 
  interface:
    viewer.viewer_name 
      # (optional static attribute) returns "official name" of this 
      # viewer class, for use in menus and such. If not defined, the classname 
      # (viewer.__name__) will be used instead. 
    viewer.is_viewable(data); 
      # (optional static method) checks if a specific data item (of the 
      # registered type) is viewable in this viewer or not. If not defined, 
      # True is assumed.
    viewer.uses_cells 
      # (optional static attribute) does this viewer uses the workspace at all?
      # True is assumed. (Should be False for viewers creating their own
      # windows).
    viewer.icon(); 
      # (optional static method) returns a QIcon for this viewer
    vo = viewer(gw,cellspec,dataitem,**opts); 
      # construct a viewer object. The gw argument is its parent 
      # GriddedWorkspace object from which the viewer may allocate cells and 
      # pages as needed. 
      # The cellspec argument should be passed to gw.allocate_cells() when 
      # allocating cell(s). This will contain information on how to allocate 
      # a cell (i.e., if an existing cell should be reused, or a new one 
      # allocated, etc.)
      # A dataitem (GridDataItem class) will be supplied at this 
      # time, although dataitem.data may be None if the object is not yet 
      # populated. 
      # **opts may be used to pass in optional keyword arguments on a 
      # per-viewer basis.
    vo.grid_page();
      # Return page on which viewer resides (a Grid.Page object), or None
      # if viewer is outside the grid.
    vo.set_data(dataitem,**opts);
      # sets/updates the content of the viewer. **opts may be used to pass 
      # in optional keyword arguments on a per-viewer basis.
    vo.add_data(dataitem,**opts); 
      # (optional method) adds another data object to the viewer. If 
      # multiple-object viewing is not supported, the viewer class should not
      # define this method. The method must return True if the object is
      # successfully added, or False if this cannot be done.
    vo.cleanup():
      # called when the viewer object is removed from the workspace
  The viewer object may also issue one Qt signal: PYSIGNAL("refresh()",(udi,)), 
  which requests a refresh of the data item given by the udi.
  Note that the GridCell interface already provides a refresh button which does 
  the same on behalf of the viewer.
  """;
  global _reg_viewers;
  global _reg_viewers_byname;
  name = getattr(viewer,'viewer_name',viewer.__name__);
  _dprint(1,"registering",name,"for type",tp);
  _reg_viewers.setdefault(tp,[]).append((priority,viewer,check_udi));
  _reg_viewers_byname[name] = viewer;

def isViewableWith (arg,viewer):
  if type(arg) is type:
    return True;
  try: checker = viewer.is_viewable;
  except AttributeError: return True; 
  if callable(checker):
    return checker(arg);
  return True;

def isViewable (arg,udi=None):
  global _reg_viewers;
  # arg may specify a type or a data object
  if type(arg) is type:
    datatype = arg;
  else:
    datatype = type(arg);
  for (tp,vlist) in _reg_viewers.items():
    # registered type must be a superclass of the supplied type;
    # registered dmi type must be either None or match the argument dmi type
    if issubclass(datatype,tp):
      for (pri,viewer,check_udi) in vlist:
        if isViewableWith(arg,viewer) and ( not udi or check_udi(udi) ):
          return True;
  return False;
  
def getViewerByName (name):
  global _reg_viewers_byname;
  return _reg_viewers_byname.get(name,None);

def getViewerList (arg,udi=None):
  global _reg_viewers;
  if arg is None:
    return [];
  # arg may specify a type or a data object
  if type(arg) is type:
    datatype = arg;
  else:
    datatype = type(arg);
  viewer_pri = {};
  # resolve data type (argument may be object or type)
  _dprint(3,udi,type(arg),'looking for viewers for',datatype);
  for (tp,vlist) in _reg_viewers.items():
    # find viewers for this class
    _dprint(3,udi,datatype,'subclass of',tp,issubclass(datatype,tp));
    if issubclass(datatype,tp):
      for (pri,v,checker) in vlist:
        _dprint(3,udi,'isViewableWith(arg,v)',v,isViewableWith(arg,v),checker(udi));
        if type(arg) is type or isViewableWith(arg,v): # if specified as object, check viewability
          # check udi if specified
          if udi is None or checker(udi):
            _dprint(3,udi,type(arg),'viewer',v,'priority',pri);
            viewer_pri[v] = min(pri,viewer_pri.get(v,999999));
  # return list sorted by priority
  vlist = list(viewer_pri.keys());
  from past.builtins import cmp
  from functools import cmp_to_key
  vlist.sort(key=cmp_to_key(lambda a,b,dd=viewer_pri:cmp(dd[a],dd[b])));
  return vlist;

class Floater (QMainWindow):
  """implements a floating window""";
  def __init__ (self,parent):
    QMainWindow.__init__(self,parent,Qt.Dialog);
    self.setWindowIcon(pixmaps.float_window.icon());
#  def hideEvent (self,ev):
#    _dprint(0,'hideEvent',ev);
#    self.emit(SIGNAL("hidden"));
  def closeEvent (self,ev):
    _dprint(2,'closeEvent',ev);
    self.emit(SIGNAL("closed"));
    QMainWindow.closeEvent(self,ev);
    
#    self.hide();
#    ev.ignore();
    
_dataitems = {};
_current_gw = None;
    
def setDefaultWorkspace (gw):
  global _current_gw;
  _current_gw = gw;
  
def getCurrentWorkspace ():
  return _current_gw;
  
def addDataItem (item,gw=None,show_gw=True,viewer=None,position=None,avoid_pos=None,newcell=False,newpage=False):
  """Adds a data cell with a viewer the given item.
       viewer:    if not None, must a be a viewer plugin class, or name. 
                  Overrides the viewer specified by the item.
       gw:        if not None, specifies a non-default gridded workspace to
                  display the item on.
       position:  if not None, must be a tuple specifying a cell to allocate, 
                  as (gridpage,x,y)
       avoid_pos: if not None, must be a tuple specifying a cell to AVOID
                  allocating, as (gridpage,x,y). 
       newpage:   if True, creates a new page with the data cell
       newcell:   if True, uses an empty cell (changing layouts as needed)
                  rather than reusing an existing unpinned panel. If False,
                  reuses a panel if possible.
       Note that the position, newpage and newcell argument are rolled up into
       a "cell specification" dict, and passed to the viewer object as the 
       cellspec argument. The cellspec is then passed as keyword arguments
       (via **cellspec) to Workspace.allocate_cells() when the viewer goes 
       to allocate cells for itself.
     Return value is either the item object itself if added, or if the
     same item (i.e. same udi/viewer) is already displayed, then the old
     item object is returned, or None on error.
  """;
  _dprint(2,item.udi,item.viewer);
  global _dataitems,_current_gw;
  # if position is specified, select workspace from page
  gw = None;
  if position is not None and isinstance(position[0],MeqGUI.Grid.Page):
    gw = position[0].gw();
  item._gridded_workspace = gw = gw or _current_gw;
  if isinstance(viewer,str):
    vc = getViewerByName(viewer,None);
    if vc:
      viewer = vc;
    else:
      raise TypeError("unknown viewer type "+viewer);
  else:  
    viewer = viewer or item.viewer;
  if not viewer:
    if not item.viewer_list:
      raise TypeError("no viewers registered and none explicitly specified");
    viewer = self.viewer_list[0];
  # Are we already displaying this item? (if not, init with empty list)
  itemlist = _dataitems.setdefault(item.udi,[]);
  # see if we can get away with not adding a viewer at all, because
  # the item is already displayed on the current page with the same viewer,
  # and we're not requested to put it into a new cell or a specific cell
  if itemlist and not (position or newcell or newpage):
    _dprint(2,'item already displayed, checking for suitable viewers');
    _dprint(2,'current page',_current_gw.current_page());
    for item0 in itemlist:
      _dprint(2,'item page',item0.viewer_obj.grid_page());
      if item0.viewer == viewer and item0.viewer_obj.grid_page() is _current_gw.current_page():
        _dprint(2,'found visible viewer');
        # in this case, highlight the data item, ask for a refresh, and
        # let the viewers take care of the rest
        highlightDataItem(item0);
        if item.data is not None:
          item0.update(item.data);
        else:
          item0.refresh();
        if show_gw:
          gw.show();
        return item0;
  # ok, we got to here so we have to create a viewer
  vopts = {};
  vopts.update(item.viewopts.get(None,{}));
  for (vclass,vo) in item.viewopts.items():
    if vclass and issubclass(viewer,vclass):
      vopts.update(vo);
  cellspec = { 'position':position,'avoid_pos':avoid_pos,'newcell':newcell,'newpage':newpage };
  try:
    item.viewer_obj = viewer(gw,dataitem=item,cellspec=cellspec,**vopts);
  except:
    (et,ev,etb) = sys.exc_info();
    _dprint(0,'error creating plugin',viewer.__name__);
    traceback.print_exc();
    errstr = '<qt><center><big>Error creating a plug-in of class <b>%s</b> for item <tt>%s</tt>.</big></center>'%(viewer.__name__,item.udi);
    errstr += '<p><tt>%s</tt>: %s'% \
        (getattr(ev,'_classname',ev.__class__.__name__),getattr(ev,'__doc__',''));
    if hasattr(ev,'args'):
      errstr += '</p><p>("'+' '.join(ev.args)+'")</p>';
    errstr += '</p></qt>';
    QMessageBox.warning(None,'Error initializing plug-in',errstr,QMessageBox.Ok,QMessageBox.NoButton);
    return None;
  # add to list
  itemlist.append(item);
  # highlight it
  highlightDataItem(item);
  # ask for a refresh of the item
  item.refresh();
  # show the workspace
  if show_gw:
    gw.show();
  return item;

def removeDataItem (item):
  global _dataitems;
  global _highlighted_item;
  itemlist = _dataitems.get(item.udi,[]);
  _dprint(2,item.udi,len(itemlist),'instances');
  # remove from highlight list
  if item is _highlighted_item:
    _highlighted_item = None;
    _current_gw.wtop().emit(SIGNAL("itemSelected"),None,);
  # remove from item list
  for i in range(len(itemlist)):
    if itemlist[i] is item:
      _dprint(2,'removing instance',i);
      del itemlist[i];
      item.cleanup();
      return;
  _dprint(2,'no matching items found');
  
_highlighted_item = None;
  
def highlightDataItem (item):
  global _highlighted_item;
  _dprint(2,item.udi);
  # remove highlights from previous cells, if any
  if _highlighted_item:
    if item is _highlighted_item:
      _dprint(3,'already highlighted');
      return;
    else:
      _dprint(3,'de-highlighting previous',_highlighted_item.udi);
      _highlighted_item.highlight(False);
  _dprint(3,'highlighting',item.udi);
  _highlighted_item = item;  
  item.highlight(True);
  _current_gw.wtop().emit(SIGNAL("itemSelected"),item,);
  
def getHighlightedItem ():
  return _highlighted_item;

# updates a data item, if it is known
def updateDataItem (udi,data):
  global _dataitems;
  _dprint(3,udi);
  _dprint(2,_dataitems);
  # scan current data items to see which need updating
  for (u,itemlist) in _dataitems.items():
    if u == udi or u.startswith(udi+'/'):
      _dprint(3,len(itemlist),'items for',u);
      update = True;
      data1 = data;  # data1 may get modified below
      subudi = u[len(udi):];
      if subudi:   # a sub-udi, so we must process it by indexing into the data
        # split into keys and process one by one (first one is empty string)
        for key in subudi.split("/")[1:]:
          try: data1 = data1[key];
          except (TypeError,KeyError): # try to convert key to integer instead
            try: data1 = data1[int(key)];
            except: 
              update = False;
              break;
      # update for all items in list
      if update:
        for item in itemlist:
          _dprint(3,'updating',item);
          item.update(data1);
